from __future__ import annotations

from typing import TypedDict

from omu.api import Extension, ExtensionType
from omu.api.endpoint import EndpointType
from omu.api.registry import RegistryType
from omu.api.registry.registry import RegistryPermissions
from omu.api.table import TablePermissions, TableType
from omu.app import App
from omu.omu import Omu

SERVER_EXTENSION_TYPE = ExtensionType("server", lambda client: ServerExtension(client))


class AppIndexEntry(TypedDict):
    url: str
    added_at: str


class AppIndex(TypedDict):
    indexes: dict[str, AppIndexEntry]


SERVER_APPS_WRITE_PERMISSION_ID = SERVER_EXTENSION_TYPE / "apps" / "write"
SERVER_APPS_READ_PERMISSION_ID = SERVER_EXTENSION_TYPE / "apps" / "read"
SERVER_INDEX_READ_PERMISSION_ID = SERVER_EXTENSION_TYPE / "index" / "read"
SERVER_INDEX_REGISTRY_TYPE = RegistryType[AppIndex].create_json(
    SERVER_EXTENSION_TYPE,
    "index",
    default_value={"indexes": {}},
    permissions=RegistryPermissions(
        write=SERVER_APPS_WRITE_PERMISSION_ID,
        read=SERVER_INDEX_READ_PERMISSION_ID,
    ),
)
SERVER_APP_TABLE_TYPE = TableType.create_model(
    SERVER_EXTENSION_TYPE,
    "apps",
    App,
    permissions=TablePermissions(
        all=SERVER_APPS_WRITE_PERMISSION_ID,
        read=SERVER_APPS_READ_PERMISSION_ID,
    ),
)
SERVER_SHUTDOWN_PERMISSION_ID = SERVER_EXTENSION_TYPE / "shutdown"
SHUTDOWN_ENDPOINT_TYPE = EndpointType[bool, bool].create_json(
    SERVER_EXTENSION_TYPE,
    "shutdown",
    permission_id=SERVER_SHUTDOWN_PERMISSION_ID,
)
TRUSTED_ORIGINS_GET_PERMISSION_ID = SERVER_EXTENSION_TYPE / "trusted_origins" / "get"
TRUSTED_ORIGINS_SET_PERMISSION_ID = SERVER_EXTENSION_TYPE / "trusted_origins" / "set"
TRUSTED_HOSTS_REGISTRY_TYPE = RegistryType[dict[str, str]].create_json(
    SERVER_EXTENSION_TYPE,
    "trusted_hosts",
    default_value={},
    permissions=RegistryPermissions(
        read=TRUSTED_ORIGINS_GET_PERMISSION_ID,
        write=TRUSTED_ORIGINS_SET_PERMISSION_ID,
    ),
)


class ServerExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return SERVER_EXTENSION_TYPE

    def __init__(self, omu: Omu) -> None:
        self._client = omu
        self.apps = omu.tables.get(SERVER_APP_TABLE_TYPE)
        self.index = omu.registries.get(SERVER_INDEX_REGISTRY_TYPE)
        self.sessions = omu.tables.get(SERVER_APP_TABLE_TYPE)
        self.trusted_origins = omu.registries.get(TRUSTED_HOSTS_REGISTRY_TYPE)

    async def shutdown(self, restart: bool = False) -> bool:
        return await self._client.endpoints.call(SHUTDOWN_ENDPOINT_TYPE, restart)



from .extension import (
    PLUGIN_MANAGE_PACKAGE_PERMISSION_ID,
    PLUGIN_READ_PACKAGE_PERMISSION_ID,
)
from .package_info import PackageInfo
from .plugin import PluginPackageInfo

__all__ = [
    "PackageInfo",
    "PluginPackageInfo",
    "PLUGIN_MANAGE_PACKAGE_PERMISSION_ID",
    "PLUGIN_READ_PACKAGE_PERMISSION_ID",
]



from dataclasses import dataclass
from typing import TypedDict

from omu.interface import Keyable
from omu.model import Model


class PluginPackageInfoData(TypedDict):
    package: str
    version: str


@dataclass(frozen=True, slots=True)
class PluginPackageInfo(Keyable, Model[PluginPackageInfoData]):
    package: str
    version: str

    @classmethod
    def from_json(cls, json: PluginPackageInfoData) -> "PluginPackageInfo":
        return cls(
            package=json["package"],
            version=json["version"],
        )

    def to_json(self) -> PluginPackageInfoData:
        return {"package": self.package, "version": self.version}

    def key(self) -> str:
        return self.package



from __future__ import annotations

import asyncio
from typing import Final

from omu import version
from omu.address import Address
from omu.app import App
from omu.event_emitter import EventEmitter, Unlisten
from omu.helper import Coro, asyncio_error_logger
from omu.network.connection import Connection, Transport
from omu.network.packet import Packet, PacketType
from omu.network.packet.packet_types import PACKET_TYPES
from omu.network.websocket_connection import WebsocketsTransport
from omu.token import JsonTokenProvider, TokenProvider


class ClientEvents:
    def __init__(self) -> None:
        self.started = EventEmitter[[]]()
        self.stopped = EventEmitter[[]]()
        self.ready = EventEmitter[[]]()


class Omu:
    def __init__(
        self,
        app: App,
        address: Address | None = None,
        token: TokenProvider | None = None,
        connection: Connection | None = None,
        transport: Transport | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
    ):
        from omu.api.asset.extension import ASSET_EXTENSION_TYPE
        from omu.api.dashboard.extension import DASHBOARD_EXTENSION_TYPE
        from omu.api.endpoint.extension import ENDPOINT_EXTENSION_TYPE
        from omu.api.i18n.extension import I18N_EXTENSION_TYPE
        from omu.api.logger.extension import LOGGER_EXTENSION_TYPE
        from omu.api.permission.extension import PERMISSION_EXTENSION_TYPE
        from omu.api.plugin.extension import PLUGIN_EXTENSION_TYPE
        from omu.api.registry.extension import REGISTRY_EXTENSION_TYPE
        from omu.api.server.extension import SERVER_EXTENSION_TYPE
        from omu.api.session.extension import SESSION_EXTENSION_TYPE
        from omu.api.signal.extension import SIGNAL_EXTENSION_TYPE
        from omu.api.table.extension import TABLE_EXTENSION_TYPE
        from omu.network import Network

        self.version: Final = version.VERSION
        self.loop = self.set_loop(loop or asyncio.new_event_loop())
        self._ready = False
        self._running = False
        self.event: Final = ClientEvents()
        self.app: Final = app
        self.address = address or Address("127.0.0.1", 26423)
        self.network: Final = Network(
            self,
            self.address,
            token or JsonTokenProvider(),
            transport=transport or WebsocketsTransport(self.address),
            connection=connection,
        )
        self.network.add_packet_handler(PACKET_TYPES.READY, self._handle_ready)

        self.sessions = SESSION_EXTENSION_TYPE.create(self)
        self.endpoints: Final = ENDPOINT_EXTENSION_TYPE.create(self)
        self.plugins: Final = PLUGIN_EXTENSION_TYPE.create(self)
        self.tables: Final = TABLE_EXTENSION_TYPE.create(self)
        self.registries: Final = REGISTRY_EXTENSION_TYPE.create(self)
        self.signals: Final = SIGNAL_EXTENSION_TYPE.create(self)
        self.permissions: Final = PERMISSION_EXTENSION_TYPE.create(self)
        self.server: Final = SERVER_EXTENSION_TYPE.create(self)
        self.assets: Final = ASSET_EXTENSION_TYPE.create(self)
        self.dashboard: Final = DASHBOARD_EXTENSION_TYPE.create(self)
        self.i18n: Final = I18N_EXTENSION_TYPE.create(self)
        self.logger: Final = LOGGER_EXTENSION_TYPE.create(self)

    async def _handle_ready(self, detail: None) -> None:
        self._ready = True
        await self.event.ready()

    def set_loop(self, loop: asyncio.AbstractEventLoop) -> asyncio.AbstractEventLoop:
        loop.set_exception_handler(asyncio_error_logger)
        self.loop = loop
        return loop

    @property
    def ready(self) -> bool:
        return self._ready

    @property
    def running(self) -> bool:
        return self._running

    async def send[T](self, type: PacketType[T], data: T) -> None:
        await self.network.send(Packet(type, data))

    def run(
        self,
        *,
        loop: asyncio.AbstractEventLoop | None = None,
        reconnect: bool = True,
    ) -> None:
        self.loop = self.set_loop(loop or self.loop)

        async def _run():
            try:
                await self.start(reconnect=reconnect)
            finally:
                if self._running:
                    await self.stop()

        if self.loop is None:
            asyncio.run(_run())
        else:
            self.loop.create_task(_run())

    async def start(self, *, reconnect: bool = True) -> None:
        current_loop = asyncio.get_event_loop()
        if current_loop is not self.loop:
            raise RuntimeError("Start must be called from the same loop")
        if self._running:
            raise RuntimeError("Already running")
        self._running = True
        await self.event.started()
        await self.network.connect(reconnect=reconnect)

    async def stop(self) -> None:
        if not self._running:
            raise RuntimeError("Not running")
        self._running = False
        await self.network.close()
        await self.event.stopped()

    def on_ready(self, coro: Coro[[], None]) -> Unlisten:
        if self._ready:
            self.loop.create_task(coro())
        return self.event.ready.listen(coro)

    async def wait_for_ready(self):
        if self._ready:
            return

        unlisten: Unlisten | None = None
        future = self.loop.create_future()

        async def _wait_for_ready():
            nonlocal unlisten
            if unlisten is not None:
                unlisten()
            future.set_result(None)

        unlisten = self.event.ready.listen(_wait_for_ready)
        await future



from __future__ import annotations

from dataclasses import dataclass
from typing import TypedDict

from omu.bytebuffer import ByteReader, ByteWriter
from omu.helper import map_optional
from omu.identifier import Identifier

from .registry import RegistryPermissions, RegistryPermissionsJSON


@dataclass(frozen=True, slots=True)
class RegistryPacket:
    id: Identifier
    value: bytes | None

    @classmethod
    def serialize(cls, item: RegistryPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        writer.write_boolean(item.value is not None)
        if item.value is not None:
            writer.write_uint8_array(item.value)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> RegistryPacket:
        with ByteReader(item) as reader:
            key = Identifier.from_key(reader.read_string())
            existing = reader.read_boolean()
            value = reader.read_uint8_array() if existing else None
        return RegistryPacket(key, value)


class RegisterPacketJSON(TypedDict):
    id: str
    permissions: RegistryPermissionsJSON


@dataclass(frozen=True, slots=True)
class RegisterPacket:
    id: Identifier
    permissions: RegistryPermissions

    @classmethod
    def serialize(cls, item: RegisterPacket) -> RegisterPacketJSON:
        return {
            "id": item.id.key(),
            "permissions": item.permissions.to_json(),
        }

    @classmethod
    def deserialize(cls, item: RegisterPacketJSON) -> RegisterPacket:
        perms = item["permissions"]
        return RegisterPacket(
            id=Identifier.from_key(item["id"]),
            permissions=RegistryPermissions(
                all=map_optional(perms.get("all"), Identifier.from_key),
                read=map_optional(perms.get("read"), Identifier.from_key),
                write=map_optional(perms.get("write"), Identifier.from_key),
            ),
        )



from __future__ import annotations

from asyncio import Future
from dataclasses import dataclass

from omu.api import Extension, ExtensionType
from omu.helper import Coro
from omu.identifier import Identifier
from omu.network.packet import PacketType
from omu.omu import Omu

from .endpoint import EndpointType
from .packets import (
    EndpointInvokedPacket,
    EndpointInvokePacket,
    EndpointRegisterPacket,
    EndpointResponsePacket,
    InvokeParams,
    ResponseParams,
)

ENDPOINT_EXTENSION_TYPE = ExtensionType("endpoint", lambda client: EndpointExtension(client))


@dataclass(frozen=True, slots=True)
class EndpointHandler:
    endpoint_type: EndpointType
    func: Coro[[EndpointInvokedPacket], bytes]


class EndpointExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return ENDPOINT_EXTENSION_TYPE

    def __init__(self, omu: Omu) -> None:
        self.omu = omu
        self.response_futures: dict[int, Future[EndpointResponsePacket]] = {}
        self.registered_endpoints: dict[Identifier, EndpointHandler] = {}
        self.call_id = 0
        omu.network.register_packet(
            ENDPOINT_REGISTER_PACKET,
            ENDPOINT_INVOKE_PACKET,
            ENDPOINT_INVOKED_PACKET,
            ENDPOINT_RESPONSE_PACKET,
        )
        omu.network.add_packet_handler(ENDPOINT_RESPONSE_PACKET, self._on_response)
        omu.network.add_packet_handler(ENDPOINT_INVOKED_PACKET, self._on_invoked)
        omu.network.add_task(self.on_ready)

    async def _on_response(self, packet: EndpointResponsePacket) -> None:
        future = self.response_futures.pop(packet.params.key, None)
        if future is None:
            raise Exception(f"Received response for unknown call key {packet.params.key} ({packet.params.id.key()})")
        future.set_result(packet)

    async def _on_invoked(self, packet: EndpointInvokedPacket) -> None:
        handler = self.registered_endpoints.get(packet.params.id)
        if handler is None:
            raise Exception(f"Received invocation for unknown endpoint {packet.params.id.key()} ({packet.params.key})")
        try:
            result = await handler.func(packet)
            await self.omu.send(
                ENDPOINT_RESPONSE_PACKET,
                EndpointResponsePacket(
                    params=ResponseParams(
                        id=packet.params.id,
                        key=packet.params.key,
                        error=None,
                    ),
                    buffer=result,
                ),
            )
        except Exception as e:
            await self.omu.send(
                ENDPOINT_RESPONSE_PACKET,
                EndpointResponsePacket(
                    params=ResponseParams(
                        id=packet.params.id,
                        key=packet.params.key,
                        error=repr(e),
                    ),
                    buffer=b"",
                ),
            )
            raise e

    async def on_ready(self) -> None:
        endpoints = {key: endpoint.endpoint_type.permission_id for key, endpoint in self.registered_endpoints.items()}
        packet = EndpointRegisterPacket(endpoints=endpoints)
        await self.omu.send(ENDPOINT_REGISTER_PACKET, packet)

    def register[Req, Res](self, type: EndpointType[Req, Res], func: Coro[[Req], Res]) -> None:
        if self.omu.ready:
            raise Exception("Cannot register endpoint after client is ready")
        if type.id in self.registered_endpoints:
            raise Exception(f"Endpoint for key {type.id} already registered")

        async def wrapper(packet: EndpointInvokedPacket) -> bytes:
            request = type.request_serializer.deserialize(packet.buffer)
            result = await func(request)
            return type.response_serializer.serialize(result)

        self.registered_endpoints[type.id] = EndpointHandler(
            endpoint_type=type,
            func=wrapper,
        )

    def bind[T, R](
        self,
        handler: Coro[[T], R] | None = None,
        endpoint_type: EndpointType[T, R] | None = None,
    ):
        if endpoint_type is None:
            raise Exception("Endpoint type must be provided")

        def decorator(func: Coro[[T], R]) -> Coro[[T], R]:
            self.register(endpoint_type, func)
            return func

        if handler:
            decorator(handler)
        return decorator

    async def call[Req, Res](self, endpoint: EndpointType[Req, Res], data: Req) -> Res:
        try:
            self.call_id += 1
            future = Future[EndpointResponsePacket]()
            self.response_futures[self.call_id] = future
            serialized = endpoint.request_serializer.serialize(data)
            await self.omu.send(
                ENDPOINT_INVOKE_PACKET,
                EndpointInvokePacket(
                    params=InvokeParams(
                        id=endpoint.id,
                        key=self.call_id,
                    ),
                    buffer=serialized,
                ),
            )
            res = await future
            return endpoint.response_serializer.deserialize(res.buffer)
        except Exception as e:
            raise Exception(f"Error calling endpoint {endpoint.id.key()}") from e


ENDPOINT_REGISTER_PACKET = PacketType[EndpointRegisterPacket].create_serialized(
    ENDPOINT_EXTENSION_TYPE,
    name="register",
    serializer=EndpointRegisterPacket,
)
ENDPOINT_INVOKE_PACKET = PacketType[EndpointInvokePacket].create_serialized(
    ENDPOINT_EXTENSION_TYPE,
    name="invoke",
    serializer=EndpointInvokePacket,
)
ENDPOINT_INVOKED_PACKET = PacketType[EndpointInvokedPacket].create_serialized(
    ENDPOINT_EXTENSION_TYPE,
    name="invoked",
    serializer=EndpointInvokedPacket,
)
ENDPOINT_RESPONSE_PACKET = PacketType[EndpointResponsePacket].create_serialized(
    ENDPOINT_EXTENSION_TYPE,
    name="response",
    serializer=EndpointResponsePacket,
)



from __future__ import annotations

import asyncio
import hashlib
import re
import typing

from loguru import logger

type Coro[**P, T] = typing.Callable[P, typing.Coroutine[None, None, T]]
type AsyncCallback[**P] = Coro[P, typing.Any]


def instance[T](cls: type[T]) -> T:
    return cls()


def map_optional[V, T](data: V | None, func: typing.Callable[[V], T], default: T | None = None) -> T | None:
    if data is None:
        return default
    return func(data)


sanitize_re = re.compile(r"[^\w]")


def sanitize_filename(name: str) -> str:
    return sanitize_re.sub("_", name)


def generate_md5_hash(id: str) -> str:
    return hashlib.md5(id.encode()).hexdigest()


def batch_call(*funcs: typing.Callable[[], None]) -> typing.Callable[[], None]:
    def wrapper():
        for func in funcs:
            func()

    return wrapper


def asyncio_error_logger(loop: asyncio.AbstractEventLoop, context: typing.Mapping) -> None:
    exception = context.get("exception")
    logger.opt(exception=exception).error("Unhandled exception")
    if exception:
        raise exception



from __future__ import annotations

import json
from collections.abc import Mapping
from dataclasses import dataclass

from omu.bytebuffer import ByteReader, ByteWriter, Flags
from omu.helper import map_optional
from omu.identifier import Identifier

from .table import TableConfig


@dataclass(frozen=True, slots=True)
class TablePacket:
    id: Identifier

    @classmethod
    def serialize(cls, item: TablePacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> TablePacket:
        with ByteReader(item) as reader:
            id = reader.read_string()
        return TablePacket(id=Identifier.from_key(id))


@dataclass(frozen=True, slots=True)
class TableItemsPacket:
    id: Identifier
    items: Mapping[str, bytes]

    @classmethod
    def serialize(cls, item: TableItemsPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        writer.write_uleb128(len(item.items))
        for key, value in item.items.items():
            writer.write_string(key)
            writer.write_uint8_array(value)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> TableItemsPacket:
        with ByteReader(item) as reader:
            id = reader.read_string()
            item_count = reader.read_uleb128()
            items: Mapping[str, bytes] = {}
            for _ in range(item_count):
                item_key = reader.read_string()
                value = reader.read_uint8_array()
                items[item_key] = value
        return TableItemsPacket(id=Identifier.from_key(id), items=items)


@dataclass(frozen=True, slots=True)
class TableKeysPacket:
    id: Identifier
    keys: tuple[str, ...]

    @classmethod
    def serialize(cls, item: TableKeysPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        writer.write_uleb128(len(item.keys))
        for key in item.keys:
            writer.write_string(key)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> TableKeysPacket:
        with ByteReader(item) as reader:
            id = reader.read_string()
            key_count = reader.read_uleb128()
            keys = [reader.read_string() for _ in range(key_count)]
        return TableKeysPacket(id=Identifier.from_key(id), keys=tuple(keys))


@dataclass(frozen=True, slots=True)
class TableProxyPacket:
    id: Identifier
    items: Mapping[str, bytes]
    key: int

    @classmethod
    def serialize(cls, item: TableProxyPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        writer.write_uleb128(item.key)
        writer.write_uleb128(len(item.items))
        for key, value in item.items.items():
            writer.write_string(key)
            writer.write_uint8_array(value)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> TableProxyPacket:
        with ByteReader(item) as reader:
            id = reader.read_string()
            key = reader.read_uleb128()
            item_count = reader.read_uleb128()
            items: Mapping[str, bytes] = {}
            for _ in range(item_count):
                item_key = reader.read_string()
                value = reader.read_uint8_array()
                items[item_key] = value
        return TableProxyPacket(
            id=Identifier.from_key(id),
            key=key,
            items=items,
        )


@dataclass(frozen=True, slots=True)
class TableFetchPacket:
    id: Identifier
    limit: int
    backward: bool
    cursor: str | None

    @classmethod
    def serialize(cls, item: TableFetchPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        writer.write_uleb128(item.limit)
        flags = Flags(length=2)
        if item.backward:
            flags.set(0)
        if item.cursor is not None:
            flags.set(1)
        writer.write_flags(flags)
        if item.cursor is not None:
            writer.write_string(item.cursor)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> TableFetchPacket:
        with ByteReader(item) as reader:
            id = reader.read_string()
            limit = reader.read_uleb128()
            flags = reader.read_flags(length=2)
            backward = flags.get(0)
            cursor = reader.read_string() if flags.has(1) else None
        return TableFetchPacket(
            id=Identifier.from_key(id),
            limit=limit,
            backward=backward,
            cursor=cursor,
        )


@dataclass(frozen=True, slots=True)
class TableFetchRangePacket:
    id: Identifier
    start: str
    end: str

    @classmethod
    def serialize(cls, item: TableFetchRangePacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        writer.write_string(item.start)
        writer.write_string(item.end)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> TableFetchRangePacket:
        with ByteReader(item) as reader:
            id = reader.read_string()
            start = reader.read_string()
            end = reader.read_string()
        return TableFetchRangePacket(id=Identifier.from_key(id), start=start, end=end)


@dataclass(frozen=True, slots=True)
class SetConfigPacket:
    id: Identifier
    config: TableConfig

    @classmethod
    def serialize(cls, item: SetConfigPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        writer.write_string(json.dumps(item.config))
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> SetConfigPacket:
        with ByteReader(item) as reader:
            id = reader.read_string()
            config = json.loads(reader.read_string())
        return SetConfigPacket(id=Identifier.from_key(id), config=config)


@dataclass(frozen=True, slots=True)
class SetPermissionPacket:
    id: Identifier
    all: Identifier | None = None
    read: Identifier | None = None
    write: Identifier | None = None
    remove: Identifier | None = None
    proxy: Identifier | None = None

    @staticmethod
    def serialize(item: SetPermissionPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        flags = 0
        if item.all is not None:
            flags |= 0b1
        if item.read is not None:
            flags |= 0b10
        if item.write is not None:
            flags |= 0b100
        if item.remove is not None:
            flags |= 0b1000
        if item.proxy is not None:
            flags |= 0b10000
        writer.write_uint8(flags)
        if item.all is not None:
            writer.write_string(item.all.key())
        if item.read is not None:
            writer.write_string(item.read.key())
        if item.write is not None:
            writer.write_string(item.write.key())
        if item.remove is not None:
            writer.write_string(item.remove.key())
        if item.proxy is not None:
            writer.write_string(item.proxy.key())
        return writer.finish()

    @staticmethod
    def deserialize(item: bytes) -> SetPermissionPacket:
        with ByteReader(item) as reader:
            id = reader.read_string()
            flags = reader.read_uint8()
            permission = reader.read_string() if flags & 0b1 else None
            permission_read = reader.read_string() if flags & 0b10 else None
            permission_write = reader.read_string() if flags & 0b100 else None
            permission_remove = reader.read_string() if flags & 0b1000 else None
            permission_proxy = reader.read_string() if flags & 0b10000 else None
        return SetPermissionPacket(
            id=Identifier.from_key(id),
            all=map_optional(permission, Identifier.from_key),
            read=map_optional(permission_read, Identifier.from_key),
            write=map_optional(permission_write, Identifier.from_key),
            remove=map_optional(permission_remove, Identifier.from_key),
            proxy=map_optional(permission_proxy, Identifier.from_key),
        )



from __future__ import annotations

from .extension import (
    SERVER_APPS_READ_PERMISSION_ID,
    SERVER_SHUTDOWN_PERMISSION_ID,
    TRUSTED_ORIGINS_GET_PERMISSION_ID,
    TRUSTED_ORIGINS_SET_PERMISSION_ID,
)

__all__ = [
    "SERVER_APPS_READ_PERMISSION_ID",
    "SERVER_SHUTDOWN_PERMISSION_ID",
    "TRUSTED_ORIGINS_GET_PERMISSION_ID",
    "TRUSTED_ORIGINS_SET_PERMISSION_ID",
]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from omu.bytebuffer import ByteReader, ByteWriter
from omu.identifier import Identifier


class LogType(Enum):
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"
    DEBUG = "debug"


@dataclass(frozen=True, slots=True)
class LogMessage:
    type: LogType
    text: str

    @classmethod
    def error(cls, text: str) -> LogMessage:
        return LogMessage(LogType.ERROR, text)

    @classmethod
    def warning(cls, text: str) -> LogMessage:
        return LogMessage(LogType.WARNING, text)

    @classmethod
    def info(cls, text: str) -> LogMessage:
        return LogMessage(LogType.INFO, text)

    @classmethod
    def debug(cls, text: str) -> LogMessage:
        return LogMessage(LogType.DEBUG, text)


@dataclass(frozen=True, slots=True)
class LogPacket:
    id: Identifier
    message: LogMessage

    @classmethod
    def serialize(cls, item: LogPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.id.key())
        writer.write_string(item.message.type.value)
        writer.write_string(item.message.text)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> LogPacket:
        with ByteReader(item) as reader:
            id = Identifier.from_key(reader.read_string())
            type = LogType(reader.read_string())
            text = reader.read_string()
            message = LogMessage(type, text)
            return LogPacket(id, message)



from __future__ import annotations

import abc
from collections.abc import Callable
from typing import Literal, NoReturn, TypeGuard, final


class UnwrapError[T, E](Exception):
    def __init__(self, result: Result[T, E], message: str):
        self.result = result
        self.message = message
        super().__init__(message)


class Monad[T, E](abc.ABC):
    @property
    @abc.abstractmethod
    def is_ok(self) -> Literal[True, False]: ...

    @property
    @abc.abstractmethod
    def is_err(self) -> Literal[True, False]: ...


@final
class Ok[T, E](Monad[T, E]):
    __match_args__ = ("value",)
    __slots__ = ("value",)

    def __init__(self, value: T):
        self.value = value

    @property
    def is_ok(self) -> Literal[True]:
        return True

    @property
    def is_err(self) -> Literal[False]:
        return False

    def unwrap(self) -> T:
        return self.value

    def apply(self, f: Callable[[T], None]) -> Result[T, E]:
        f(self.value)
        return self

    def apply_err(self, f: Callable[[E], None]) -> Result[T, E]:
        return self

    def map[V](self, f: Callable[[T], V]) -> Result[V, E]:
        return Ok(f(self.value))

    def map_err[V](self, f: Callable[[E], V]) -> Ok[T, E]:
        return self

    def and_then[V](self, f: Callable[[T], Result[V, E]]) -> Result[V, E]:
        return f(self.value)

    def join(self, other: Result[T, E]) -> Result[T, E]:
        return other

    __and__ = join

    def join_err(self, other: Result[T, E]) -> Result[T, E]:
        return self


@final
class Err[T, E](Monad[T, E]):
    __match_args__ = ("err",)
    __slots__ = ("err",)

    def __init__(self, err: E):
        self.err = err

    @property
    def is_ok(self) -> Literal[False]:
        return False

    @property
    def is_err(self) -> Literal[True]:
        return True

    def unwrap(self) -> NoReturn:
        exc = UnwrapError(
            result=self,
            message=f"called `unwrap` on an `Err` value: {self.err}",
        )
        if isinstance(self.err, BaseException):
            raise exc from self.err
        raise exc

    def apply(self, f: Callable[[T], None]) -> Result[T, E]:
        return self

    def apply_err(self, f: Callable[[E], None]) -> Result[T, E]:
        f(self.err)
        return self

    def map[V](self, f: Callable[[T], V]) -> Result[T, E]:
        return self

    def map_err[V](self, f: Callable[[E], V]) -> Err[T, V]:
        return Err(f(self.err))

    def and_then[V](self, f: Callable[[T], Result[V, E]]) -> Result[T, E]:
        return self

    def join(self, other: Result[T, E]) -> Result[T, E]:
        return self

    __and__ = join

    def join_err(self, other: Result[T, E]) -> Result[T, E]:
        return other


type Result[T, E] = Ok[T, E] | Err[T, E]


def is_ok[T, E](result: Result[T, E]) -> TypeGuard[Ok[T, E]]:
    return result.is_ok


def is_err[T, E](result: Result[T, E]) -> TypeGuard[Err[T, E]]:
    return result.is_err



from __future__ import annotations

from omu.api import Extension, ExtensionType
from omu.event_emitter import Unlisten
from omu.helper import AsyncCallback
from omu.identifier import Identifier
from omu.network.packet import PacketType
from omu.omu import Omu
from omu.serializer import Serializer

from .packets import LogMessage, LogPacket

LOGGER_EXTENSION_TYPE = ExtensionType("logger", lambda client: LoggerExtension(client))
LOGGER_LOG_PERMISSION_ID = LOGGER_EXTENSION_TYPE / "log"

LOGGER_LOG_PACKET = PacketType[LogPacket].create_serialized(
    identifier=LOGGER_EXTENSION_TYPE,
    name="log",
    serializer=LogPacket,
)
LOGGER_LISTEN_PACKET = PacketType[Identifier].create_json(
    identifier=LOGGER_EXTENSION_TYPE,
    name="listen",
    serializer=Serializer.model(Identifier),
)
LOGGER_SERVER_ID = LOGGER_EXTENSION_TYPE / "server"


class LoggerExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return LOGGER_EXTENSION_TYPE

    def __init__(self, omu: Omu):
        omu.network.register_packet(
            LOGGER_LOG_PACKET,
            LOGGER_LISTEN_PACKET,
        )
        omu.network.add_packet_handler(LOGGER_LOG_PACKET, self.handle_log)
        omu.permissions.require(LOGGER_LOG_PERMISSION_ID)
        self.client = omu
        self.listeners: dict[Identifier, set[AsyncCallback[LogMessage]]] = {}

    async def log(self, message: LogMessage) -> None:
        packet = LogPacket(
            id=self.client.app.id,
            message=message,
        )
        await self.client.send(LOGGER_LOG_PACKET, packet)

    async def error(self, text: str) -> None:
        await self.log(LogMessage.error(text))

    async def warning(self, text: str) -> None:
        await self.log(LogMessage.warning(text))

    async def info(self, text: str) -> None:
        await self.log(LogMessage.info(text))

    async def debug(self, text: str) -> None:
        await self.log(LogMessage.debug(text))

    async def listen(self, id: Identifier, listener: AsyncCallback[[LogMessage]]) -> Unlisten:
        if id not in self.listeners:

            async def on_ready():
                await self.client.send(LOGGER_LISTEN_PACKET, id)

            self.client.on_ready(on_ready)

            self.listeners[id] = set()
        self.listeners[id].add(listener)

        return lambda: self.listeners[id].remove(listener)

    async def handle_log(self, packet: LogPacket) -> None:
        for callback in self.listeners.get(packet.id, []):
            await callback(packet.message)



from __future__ import annotations

import abc
import json
from pathlib import Path

from omu.address import Address
from omu.app import App


class TokenProvider(abc.ABC):
    @abc.abstractmethod
    def get(self, address: Address, app: App) -> str | None:
        pass

    @abc.abstractmethod
    def store(self, address: Address, app: App, token: str) -> None:
        pass


class JsonTokenProvider(TokenProvider):
    def __init__(self, path: Path | None = None):
        path = path or Path.cwd() / ".omu_cache" / "tokens.json"
        path.parent.mkdir(parents=True, exist_ok=True)
        self._path = path

    @classmethod
    def get_store_key(cls, address: Address, app: App) -> str:
        return f"{address.host}:{address.port}:{address.hash or ""}:{app.id.key()}"

    def get(self, address: Address, app: App) -> str | None:
        if not self._path.exists():
            return None
        tokens = json.loads(self._path.read_text(encoding="utf-8"))
        return tokens.get(self.get_store_key(address, app))

    def store(self, address: Address, app: App, token: str) -> None:
        tokens: dict[str, str] = {}
        if self._path.exists():
            tokens = json.loads(self._path.read_text(encoding="utf-8"))

        tokens[self.get_store_key(address, app)] = token
        self._path.write_text(json.dumps(tokens), encoding="utf-8")



from __future__ import annotations

import abc

from omu.errors import NetworkError
from omu.network.packet import Packet, PacketData, PacketType
from omu.result import Err, Ok, Result
from omu.serializer import Serializable


class CloseError(NetworkError): ...


class ConnectionClosed(NetworkError): ...


class ErrorReceiving(NetworkError): ...


class InvalidPacket(NetworkError): ...


type ReceiveError = ConnectionClosed | ErrorReceiving | InvalidPacket


class Transport(abc.ABC):
    @abc.abstractmethod
    async def connect(self) -> Connection: ...


class Connection(abc.ABC):
    @abc.abstractmethod
    async def send(
        self,
        packet: Packet,
        packet_mapper: Serializable[Packet, PacketData],
    ) -> None: ...

    @abc.abstractmethod
    async def receive(
        self,
        packet_mapper: Serializable[Packet, PacketData],
    ) -> Result[Packet, ReceiveError]: ...

    async def receive_as[T](
        self,
        packet_mapper: Serializable[Packet, PacketData],
        packet_type: PacketType[T],
    ) -> Result[T, ReceiveError]:
        packet = await self.receive(packet_mapper)
        if packet.is_err is True:
            return Err(packet.err)
        packet = packet.value
        if packet.type != packet_type:
            return Err(InvalidPacket(f"Expected {packet_type.id} but got {packet.type}"))
        return Ok(packet.data)

    @abc.abstractmethod
    async def close(self) -> None: ...

    @property
    @abc.abstractmethod
    def closed(self) -> bool: ...
